#include <SDL2/SDL.h>
#include <stdio.h>
#include "renderer.h"
#include "../microui/microui.h"
#include <stdint.h>
#include <stdbool.h>
#include <assert.h>

// I had to change -std to gnu11 instead of c11 for this one.
// Defining the posix version macro didn't do anything at all.
#include <time.h>
#include <errno.h>

#include "cpu.c"

int load_program(char *filepath) {
	bool success = true;
	FILE *file = NULL;
	if(success) {
		file = fopen(filepath, "r");
		if(file == NULL) {
			perror("Error opening file for reading");
			success = false;
	}}

	if(success) {
		int seek_status = fseek(file, 0, SEEK_END);
		if(seek_status != 0) {
			perror("Error seeking in file");
			success = false;
		}
	}

	long file_size = 0;
	if(success) {
		file_size = ftell(file);
		int seek_status = fseek(file, 0, SEEK_SET);
		if(seek_status != 0) {
			perror("Error seeking in file");
			success = false;
		}
	}

	char *data = NULL;
	if(success) {
		data = malloc(file_size);
		if(data == NULL) {
			success = false;
		}
	}

	if(success) {
		size_t read_count = fread(data, 1, file_size, file);
		if(read_count != file_size) {
			fprintf(stderr, "Incomplete read, aborting.\n");
			return 1;
		}
	}

	if(success) {
		cpu_memory_clear();
		int memory_cursor = 0;
		Cpu_Memory_Cell cell = {"00"};
		bool first_nibble = true;
		for(int i = 0; i < file_size; i++) {
			char c = data[i];
			if((c < '0') || (c > '9' && c < 'A') || (c > 'F')) {
				if(c != ' ' && c != '\n' && c != '\t') {
					fprintf(stderr, "Strange character in memory file: %c\n", c);
					success = false;
					break;
				}
			} else {
				if(memory_cursor >= NUM_MEMORY_CELLS) {
					fprintf(stderr, "Memory in file too big for machine memory.\n");
					success = false;
					break;
				}
				if(first_nibble) {
					cell.text[0] = c;
					first_nibble = false;
				} else {
					cell.text[1] = c;
					first_nibble = true;
					memory[memory_cursor] = cell;
					memory_cursor += 1;
				}
			}
		}
		if(!first_nibble) {
			fprintf(stderr, "Number of nibbles in file is odd!\n");
			success = false;
		}
	}

	if(file != NULL) {
		fclose(file);
	}
	if(data != NULL) {
		free(data);
	}
	return success ? 0 : 1;
}

int save_program(char *filepath) {
	bool success = true;
	FILE *file = NULL;
	if(success) {
		file = fopen(filepath, "w");
		if(file == NULL) {
			perror("Error opening file for writing");
			success = false;
	}}

	if(success) {
		char text[NUM_MEMORY_CELLS * 3];
		for(int i = 0; i < NUM_MEMORY_CELLS; i++) {
			char *c1 = &text[i*3];
			char *c2 = c1+1;
			char *whitespace = c2+1;
			*c1 = memory[i].text[0];
			*c2 = memory[i].text[1];
			*whitespace = (i%4==3) ? '\n' : ' ';
		}

		size_t written = fwrite(text, sizeof(text), 1, file);
		if(written == 0) {
			perror("Write error");
			success = false;
		}
	}

	if(file != NULL) {
		fclose(file);
	}
	return success ? 0 : 1;
}

int width = 800;
int height = 600;

typedef int64_t i64;



int mu_textbox_raw2(mu_Context *ctx, char *buf, int bufsz, mu_Id id, mu_Rect r,
  int opt)
{
  int res = 0;
  mu_update_control(ctx, id, r, opt | MU_OPT_HOLDFOCUS);

  if (ctx->focus == id) {
    /* handle text input */
    int len = strlen(buf);
    int n = mu_min(bufsz - len - 1, (int) strlen(ctx->input_text));
    if (n > 0) {
			char input_char = ctx->input_text[0];
			if(input_char >= 'a' && input_char <= 'f') {
				input_char -= 0x20; // Convert to upper case.
			}
			if((input_char >= '0' && input_char <= '9') || (input_char >= 'A' && input_char <= 'F'))
			{
				buf[len] = input_char;
				len++;
				buf[len] = '\0';
				res |= MU_RES_CHANGE;
			}
    }
    /* handle backspace */
    if (ctx->key_pressed & MU_KEY_BACKSPACE && len > 0) {
			--len;
      buf[len] = '\0';
      res |= MU_RES_CHANGE;
    }
    /* handle return */
    if (ctx->key_pressed & MU_KEY_RETURN) {
      mu_set_focus(ctx, 0);
      res |= MU_RES_SUBMIT;
    }
  }

  /* draw */
  mu_draw_control_frame(ctx, id, r, MU_COLOR_BASE, opt);
  if (ctx->focus == id) {
    mu_Color color = ctx->style->colors[MU_COLOR_TEXT];
    mu_Font font = ctx->style->font;
    int textw = ctx->text_width(font, buf, -1);
    int texth = ctx->text_height(font);
    int ofx = r.w - ctx->style->padding - textw - 1;
    int textx = r.x + mu_min(ofx, ctx->style->padding);
    int texty = r.y + (r.h - texth) / 2;
    mu_push_clip_rect(ctx, r);
    mu_draw_text(ctx, font, buf, -1, mu_vec2(textx, texty), color);
    mu_draw_rect(ctx, mu_rect(textx + textw, texty, 1, texth), color);
    mu_pop_clip_rect(ctx);
  } else {
    mu_draw_control_text(ctx, buf, r, MU_COLOR_TEXT, opt);
  }

  return res;
}

int mu_textbox_ex2(mu_Context *ctx, char *buf, int bufsz, int opt) {
  mu_Id id = mu_get_id(ctx, &buf, sizeof(buf));
  mu_Rect r = mu_layout_next(ctx);
  return mu_textbox_raw2(ctx, buf, bufsz, id, r, opt);
}
#define mu_textbox2(ctx, buf, bufsz)       mu_textbox_ex2(ctx, buf, bufsz, 0)

int measure_text(mu_Context *ctx, char *text)
{
	mu_Font font = ctx->style->font;
	return ctx->text_width(font, text, -1);
}


//static float bg[3] = { 90, 95, 100 };
static float bg[3] = { 255, 230, 230 };

char filepath_buffer[1024*4] = {0};

static void process_frame(
	mu_Context *ctx,
	i64 current_seconds_timestamp
) {
  mu_begin(ctx);
	if (
		//mu_begin_window(ctx, "My Window", mu_rect(10, 10, 300, 400))
		mu_begin_window_ex(ctx, "My Window", mu_rect(0, 0, width, height),
			MU_OPT_NOTITLE|MU_OPT_NORESIZE)
		) {

		mu_layout_row(ctx, 4, (int[]) { 200, 200, 200, -1 }, -1);
		mu_begin_panel(ctx, "register panel");
		mu_layout_row(ctx, 2, (int[]) { 120, -1 }, 0);
		mu_label(ctx, cpu_running ? "Running" : "Halted"); mu_label(ctx, "");
		mu_label(ctx, "Program counter:");
		mu_label(ctx, program_counter.text);
		//mu_label(ctx, "Register 0:");
		//mu_label(ctx, "0x00");
		//mu_label(ctx, "Register 1:");
		//mu_label(ctx, "0x00");
		for(int i = 0; i < NUM_REGISTERS; i++) {
			mu_label(ctx, register_names[i]);
			mu_label(ctx, registers[i].text);
		}
		mu_end_panel(ctx);

		mu_begin_panel(ctx, "memory panel");
		mu_Container *panel = mu_get_current_container(ctx);
		// NOTE(Zelaven): Garbage formula that kinda works.
#if 0
		int width = (panel->rect.w-ctx->style->padding*2-ctx->style->spacing*4)/4;
		printf("%d = %d / 4\n", width, panel->rect.w-ctx->style->padding*2-ctx->style->spacing*3);
		printf("%f = %f / 4.0\n",
			((float)panel->rect.w-(float)ctx->style->padding*2.0f-(float)ctx->style->spacing*3.0f)/4.0f,
			(float)panel->rect.w-(float)ctx->style->spacing*2.0f-(float)ctx->style->spacing*3.0f);
		mu_layout_row(ctx, 4, (int[]) { width, width, width, width }, 0);
		static char mem1[3] = {0};
		mu_textbox(ctx, mem1, sizeof(mem1));
		static char mem2[3] = {0};
		mu_textbox2(ctx, mem2, sizeof(mem2));
		static char mem3[3] = {0};
		mu_textbox2(ctx, mem3, sizeof(mem3));
		static char mem4[3] = {0};
		mu_textbox2(ctx, mem4, sizeof(mem4));
#endif
#if 0
		int marker_width = measure_text(ctx, ">__");
		int label_width = measure_text(ctx, "0x0000__");
		int cell_real_estate = (panel->rect.w-ctx->style->padding*2-ctx->style->spacing*5) - marker_width - label_width;
		int cell_width = cell_real_estate / 4;
		mu_layout_row(ctx, 6, (int[]) { marker_width, label_width, cell_width, cell_width, cell_width, cell_width }, 0);
		mu_label(ctx, ">");
		mu_label(ctx, "0x0000");
		static char mem1[3] = {0};
		mu_textbox(ctx, mem1, sizeof(mem1));
		static char mem2[3] = {0};
		mu_textbox2(ctx, mem2, sizeof(mem2));
		static char mem3[3] = {0};
		mu_textbox2(ctx, mem3, sizeof(mem3));
		static char mem4[3] = {0};
		mu_textbox2(ctx, mem4, sizeof(mem4));
		mu_end_panel(ctx);
#endif
		int marker_width = measure_text(ctx, ">__");
		int label_width = measure_text(ctx, "0x0000__");
		int cell_real_estate = (panel->rect.w-ctx->style->padding*2-ctx->style->spacing*5) - marker_width - label_width - ctx->style->scrollbar_size;
		int cells_per_row = 4;
		int cell_width = cell_real_estate / cells_per_row;
		mu_layout_row(ctx, 2+cells_per_row, (int[]) { marker_width, label_width, cell_width, cell_width, cell_width, cell_width }, 0);
		for(int i = 0; i < NUM_MEMORY_CELLS; i += cells_per_row)
		{
			mu_label(ctx, (cell_to_value(program_counter) == i) ? ">" : "");
			char buf[7] = "0x0000";
			snprintf(buf, sizeof(buf), "0x%4.4X", i);
			mu_label(ctx, buf);
			for(int j = 0; j < cells_per_row; j++) {
				mu_textbox2(ctx, memory[i+j].text, 3);
			}
		}
		mu_end_panel(ctx);

		mu_begin_panel(ctx, "buttons panel");
		mu_layout_row(ctx, 2, (int[]) { 100, -1 }, 0);
		if(mu_button(ctx, "Reset")) {
			cpu_reset();
		}
		if(mu_button(ctx, "Clear")) {
			cpu_memory_clear();
			cpu_running = false;
		}
		if(mu_button(ctx, "Step")) {
			cpu_step();
		}
		mu_end_panel(ctx);

		mu_begin_panel(ctx, "save/load panel");
		mu_layout_row(ctx, 1, (int[]) { -1 }, 0);
		mu_textbox(ctx, filepath_buffer, sizeof(filepath_buffer));
		if(mu_button(ctx, "Save")) {
			save_program(filepath_buffer);
		}
		if(mu_button(ctx, "Load")) {
			load_program(filepath_buffer);
			cpu_reset();
			cpu_running = false;
		}
		mu_end_panel(ctx);

		mu_end_window(ctx);
	}
  mu_end(ctx);
}

static const char button_map[256] = {
  [ SDL_BUTTON_LEFT   & 0xff ] =  MU_MOUSE_LEFT,
  [ SDL_BUTTON_RIGHT  & 0xff ] =  MU_MOUSE_RIGHT,
  [ SDL_BUTTON_MIDDLE & 0xff ] =  MU_MOUSE_MIDDLE,
};

static const char key_map[256] = {
  [ SDLK_LSHIFT       & 0xff ] = MU_KEY_SHIFT,
  [ SDLK_RSHIFT       & 0xff ] = MU_KEY_SHIFT,
  [ SDLK_LCTRL        & 0xff ] = MU_KEY_CTRL,
  [ SDLK_RCTRL        & 0xff ] = MU_KEY_CTRL,
  [ SDLK_LALT         & 0xff ] = MU_KEY_ALT,
  [ SDLK_RALT         & 0xff ] = MU_KEY_ALT,
  [ SDLK_RETURN       & 0xff ] = MU_KEY_RETURN,
  [ SDLK_BACKSPACE    & 0xff ] = MU_KEY_BACKSPACE,
};

static int text_width(mu_Font font, const char *text, int len) {
  if (len == -1) { len = strlen(text); }
  return r_get_text_width(text, len);
}

static int text_height(mu_Font font) {
  return r_get_text_height();
}

int main(int argc, char **argv) {
	if(argc > 1) {
		int filepath_length = strlen(argv[1]);
		int buffer_length = sizeof(filepath_buffer);
		if(filepath_length < buffer_length) {
			strncpy(filepath_buffer, argv[1], buffer_length);
		} else {
			strncpy(filepath_buffer, "Filepath is too long.", buffer_length);
		}
		if(load_program(argv[1]) != 0) {
			return 1;
		}
	} else {
		cpu_memory_clear();
	}

  /* init SDL and renderer */
  SDL_Init(SDL_INIT_EVERYTHING);
  r_init(width, height);

  /* init microui */
  mu_Context *ctx = malloc(sizeof(mu_Context));
  mu_init(ctx);
  ctx->text_width = text_width;
  ctx->text_height = text_height;

  /* main loop */
  for (;;) {
    /* handle SDL events */
    SDL_Event e;
    while (SDL_PollEvent(&e)) {
      switch (e.type) {
        case SDL_QUIT: exit(EXIT_SUCCESS); break;
        case SDL_MOUSEMOTION: mu_input_mousemove(ctx, e.motion.x, e.motion.y); break;
        case SDL_MOUSEWHEEL: mu_input_scroll(ctx, 0, e.wheel.y * -30); break;
        case SDL_TEXTINPUT: mu_input_text(ctx, e.text.text); break;

        case SDL_MOUSEBUTTONDOWN:
        case SDL_MOUSEBUTTONUP: {
          int b = button_map[e.button.button & 0xff];
          if (b && e.type == SDL_MOUSEBUTTONDOWN) { mu_input_mousedown(ctx, e.button.x, e.button.y, b); }
          if (b && e.type ==   SDL_MOUSEBUTTONUP) { mu_input_mouseup(ctx, e.button.x, e.button.y, b);   }
          break;
        }

        case SDL_KEYDOWN:
        case SDL_KEYUP: {
          int c = key_map[e.key.keysym.sym & 0xff];
          if (c && e.type == SDL_KEYDOWN) { mu_input_keydown(ctx, c); }
          if (c && e.type ==   SDL_KEYUP) { mu_input_keyup(ctx, c);   }
          break;
        }
      }
    }

		i64 current_seconds_timestamp = 0;
		{
			struct timespec clock_timestamp;
			int clock_retval = clock_gettime(CLOCK_BOOTTIME, &clock_timestamp);
			if(clock_retval != 0) {
				int error = errno;
				perror("clock_gettime failed: ");
				fprintf(stderr, "Errno number: %d\n", error);
				return 1;
			}
			current_seconds_timestamp = clock_timestamp.tv_sec;
		}

    /* process frame */
    process_frame(ctx, current_seconds_timestamp);

    /* render */
    r_clear(mu_color(bg[0], bg[1], bg[2], 255));
    mu_Command *cmd = NULL;
    while (mu_next_command(ctx, &cmd)) {
      switch (cmd->type) {
        case MU_COMMAND_TEXT: r_draw_text(cmd->text.str, cmd->text.pos, cmd->text.color); break;
        case MU_COMMAND_RECT: r_draw_rect(cmd->rect.rect, cmd->rect.color); break;
        case MU_COMMAND_ICON: r_draw_icon(cmd->icon.id, cmd->icon.rect, cmd->icon.color); break;
        case MU_COMMAND_CLIP: r_set_clip_rect(cmd->clip.rect); break;
      }
    }
    r_present();
  }

  return 0;
}

